import os
from contextlib import nullcontext

from django.core.management import call_command


def pytest_addoption(parser):
    group = parser.getgroup("migrateci")
    group.addoption(
        "--migrateci",
        action="store_true",
        help="run migrateci before tests",
    )
    group.addoption("--migrateci-location")
    group.addoption("--migrateci-storage")
    group.addoption("--migrateci-depth", type=int)
    group.addoption("--migrateci-verbose", type=int)


def pytest_configure(config):
    if config.option.help or not config.option.migrateci:
        return

    # Make pytest-django never create db because migrateci already do it.
    # It must be changed even inside the xdist worker.
    create_db = config.option.create_db
    config.option.create_db = False

    # I want to execute only on controller (without worker_id) to avoid multiple runs.
    # It also works without pytest-xdist.
    worker_id = os.environ.get("PYTEST_XDIST_WORKER")
    if worker_id is not None:
        return

    try:
        from pytest_django import plugin

        db_unblock = config.stash[plugin.blocking_manager_key].unblock
    except AttributeError:
        # pytest-django<4.7 doesn't have `blocking_manager_key` defined,
        # raising `AttributeError`.
        db_unblock = plugin._blocking_manager.unblock
    except ImportError:
        # pytest-django lib is not installed, do nothing and hope for the best.
        db_unblock = nullcontext

    verbosity = config.option.migrateci_verbose
    if verbosity is None:
        verbosity = config.option.verbose

    command_kwargs = {"pytest": True}

    # Option numprocesses is from pytest-xdist and doesn't exist if it is not installed.
    parallel = getattr(config.option, "numprocesses", None)
    if parallel:
        command_kwargs["parallel"] = parallel

    # Option reuse_db is from pytest-django.
    reuse_db = getattr(config.option, "reuse_db", False)
    if reuse_db and not create_db:
        command_kwargs["reuse_db"] = True

    if config.option.migrateci_location:
        command_kwargs["location"] = config.option.migrateci_location
    if config.option.migrateci_storage:
        command_kwargs["storage_class"] = config.option.migrateci_storage
    if config.option.migrateci_depth is not None:
        command_kwargs["depth"] = config.option.migrateci_depth
    if verbosity is not None:
        command_kwargs["verbosity"] = verbosity

    with db_unblock():
        call_command("migrateci", **command_kwargs)



import hashlib
import importlib
import itertools
import logging
import os
import re
import tempfile
from contextlib import contextmanager

from django.conf import settings
from django.db import connections
from django.db.migrations.loader import MigrationLoader
from django.test.utils import setup_databases

logger = logging.getLogger(__name__)


def _get_db_backend(connection):
    vendor_map = {
        "mysql": "django_migrations_ci.backends.mysql",
        "postgresql": "django_migrations_ci.backends.postgresql",
        "sqlite": "django_migrations_ci.backends.sqlite",
    }
    return importlib.import_module(vendor_map[connection.vendor])


def create_test_db(connection, *, keepdb=False, verbosity=1):
    backend = _get_db_backend(connection)
    test_database_name = connection.creation._get_test_db_name()
    database_created = not keepdb or not backend.database_exists(
        connection, test_database_name
    )
    database_name = connection.creation._create_test_db(
        verbosity=verbosity,
        autoclobber=True,
        keepdb=keepdb,
    )
    return database_name, database_created


def get_unique_connections():
    return [
        connection
        for connection in connections.all()
        if not connection.settings_dict.get("TEST", {}).get("MIRROR")
    ]


def setup_test_db(*, verbosity=1):
    # Based on https://github.com/django/django/blob/d62563cbb194c420f242bfced52b37d6638e67c6/django/test/runner.py#L1051-L1054
    unique_connections = get_unique_connections()
    database_names = {
        connection.alias: connection.settings_dict["NAME"]
        for connection in unique_connections
    }
    aliases = list(database_names.keys())
    setup_databases(
        aliases=aliases,
        verbosity=verbosity,
        interactive=False,
        keepdb=True,
    )

    # Django setup_databases change original settings and don't care about it
    # because it run the setup only one time and other parts of testing understand that.
    for connection in unique_connections:
        database_name = database_names[connection.alias]
        connection.close()
        settings.DATABASES[connection.alias]["NAME"] = database_name
        connection.settings_dict["NAME"] = database_name


def clone_test_db(connection, parallel, is_pytest=False, *, verbosity=1):
    for index in range(parallel):
        if is_pytest:
            # pytest-django use test_db_gwN, from 0 to N-1.
            # e.g. test_db_gw0, test_db_gw1, ...
            # https://github.com/pytest-dev/pytest-django/blob/e0c77b391ea54c3b8d6ffbb593aa25188a0ce7e9/pytest_django/fixtures.py#L61
            suffix = f"gw{index}"
        else:
            # Django use test_db_N, from 1 to N.
            # e.g. test_db_1, test_db_2, ...
            suffix = f"{index + 1}"

        connection.creation.clone_test_db(
            suffix=suffix,
            verbosity=verbosity,
            keepdb=False,
        )

        if connection.vendor == "sqlite":
            settings_dict = connection.creation.get_test_db_clone_settings(suffix)
            django_db_name = settings_dict["NAME"]
            if is_pytest and "." in django_db_name:
                # Move db_gw0.sqlite3 to db.sqlite3_gw0.
                os.rename(django_db_name, _fix_sqlite_pytest_suffix(django_db_name))


def _fix_sqlite_pytest_suffix(db_name):
    # Django<4 generate sqlite3 names with two dots, like db_1..sqlite3,
    # it is cleaned here to db_1.sqlite3.
    clean_db_name = re.sub(r"\.+", ".", db_name)
    # Django clone_test_db create file db_gw0.sqlite3, but pytest-django
    # expects db.sqlite3_gw0. Lets rename the file.
    pytest_db_name = re.sub(r"(_gw\d+)(\.)+(.+)$", r".\3\1", clean_db_name)
    return pytest_db_name  # noqa: RET504


@contextmanager
def test_db(connection, suffix=""):
    # Django clone_test_db trust setup_databases already changed original settings,
    # so I have to do that here.
    try:
        test_db_name = connection.settings_dict["TEST"]["NAME"]
    except KeyError:
        test_db_name = None

    if not test_db_name:
        test_db_name = connection.creation._get_test_db_name()

    if suffix:
        if connection.vendor == "sqlite":
            settings_dict = connection.creation.get_test_db_clone_settings(suffix)
            _generated_db_name = settings_dict["NAME"]
            # Django<4 generate sqlite3 names with two dots, like db_1..sqlite3,
            # or if it does not have an extension, database name will end with a dot.
            dotbug = ".." in _generated_db_name or _generated_db_name.endswith(".")
            test_db_name = _transform_sqlite_db_name(
                test_db_name,
                suffix=suffix,
                dotbug=dotbug,
            )
        else:
            test_db_name += f"_{suffix}"

    db_name = connection.settings_dict["NAME"]
    connection.settings_dict["NAME"] = test_db_name
    settings.DATABASES[connection.alias]["NAME"] = test_db_name
    connection.close()

    try:
        yield
    finally:
        connection.settings_dict["NAME"] = db_name
        settings.DATABASES[connection.alias]["NAME"] = db_name
        connection.close()


def _transform_sqlite_db_name(db_name, *, suffix="", dotbug=False):
    if suffix:
        suffix = f"_{suffix}"
    if dotbug:
        suffix += "."

    # db.sqlite3_1 to db_1.sqlite3.
    db_name = re.sub(r"(.+)(\..+)$", rf"\1{suffix}\2", db_name)
    # db_gw1.sqlite3 to db.sqlite3_gw1
    db_name = re.sub(r"(_gw\d+)\.+(.+)$", r".\2\1", db_name)
    return db_name  # noqa: RET504


def load(connection, input_file, storage, *, verbosity=1):
    if verbosity:
        db_name = connection.settings_dict["NAME"]
        logger.info(f"Load {input_file} to {db_name}")
    with storage.open(input_file, "r") as f:
        sql = f.read()
    with connection.cursor() as cursor:
        if connection.vendor == "sqlite":
            # sqlite3 can't use execute() to run many statements, it fails with
            # sqlite3.Warning: You can only execute one statement at a time.
            cursor.executescript(sql)
        else:
            cursor.execute(sql)


def dump(connection, output_file, storage, *, verbosity=1):
    if verbosity:
        db_name = connection.settings_dict["NAME"]
        logger.info(f"Dump {db_name} to {output_file}.")

    backend = _get_db_backend(connection)

    # Dump to a temp file because backends expect a filename instead of a file object.
    with tempfile.NamedTemporaryFile(prefix="migrateci", suffix=".sql") as tmp_sql_fp:
        backend.dump(connection, tmp_sql_fp.name)
        tmp_sql_fp.flush()

        # Copy it to the storage.
        with open(tmp_sql_fp.name, "r") as input_fp, storage.open(output_file, "w") as output_fp:
            output_fp.write(input_fp.read())


def hash_files(depth=0):
    """Generate checksums based on migrations graph.

    Yield many checksums first based on all migrations, after that it tries to
    remove `depth` migrations per app.
    """
    loader = MigrationLoader(None, ignore_no_migrations=True)
    nodes = loader.graph.leaf_nodes()
    plan = loader.graph._generate_plan(nodes, at_end=True)

    checksums = {}
    for app_label, migration_name in plan:
        migration = loader.get_migration(app_label, migration_name)
        module = importlib.import_module(migration.__module__)
        filename = module.__file__

        # Calculate checksum for each migration to limit memory usage.
        with open(filename, "rb") as f:
            checksum = hashlib.md5(f.read()).hexdigest()

        app_checksums = checksums.setdefault(app_label, [])
        app_checksums.append(checksum)

    for n in range(depth + 1):
        combinations = itertools.combinations_with_replacement(checksums, n)
        for apps_to_remove_migration in combinations:
            copy_checksums = checksums.copy()
            for app_label in apps_to_remove_migration:
                # Remove last migration from selected app.
                remaining_checksums = copy_checksums[app_label][:-1]
                copy_checksums[app_label] = remaining_checksums

            checksum = hashlib.md5()
            # Sort apps to be deterministic.
            # Django enforce migration dependencies ordering, but it is not
            # deterministic for migrations without direct dependency.
            for _, app_checksums in sorted(copy_checksums.items()):
                checksum.update("".join(app_checksums).encode())
            yield checksum.hexdigest()



from django.db.backends.utils import strip_quotes

from django_migrations_ci import shell


def dump(connection, output_file):
    ctx, env = _ctx(connection.settings_dict)
    if ctx["port"]:
        mysqldump = "mysqldump -h {host} -P {port} -u {user} --result-file {output_file} {database}"  # noqa: E501
    else:
        mysqldump = (
            "mysqldump -h {host} -u {user} --result-file {output_file} {database}"
        )
    shell.exec(mysqldump.format(output_file=output_file, **ctx), env)


def database_exists(connection, database_name):
    with connection.creation._nodb_cursor() as cursor:
        cursor.execute(
            "SELECT 1 FROM information_schema.schemata WHERE schema_name = %s",
            [strip_quotes(database_name)],
        )
        return cursor.fetchone() is not None


def _ctx(db_conf):
    env = {"MYSQL_PWD": db_conf["PASSWORD"]}
    try:
        database = db_conf["TEST"]["NAME"] or db_conf["NAME"]
    except KeyError:
        database = db_conf["NAME"]

    data = {
        "host": db_conf["HOST"],
        "port": db_conf["PORT"],
        "user": db_conf["USER"],
        "database": database,
    }
    return data, env



import os

from django_migrations_ci import shell


def dump(connection, output_file):
    database = connection.settings_dict["NAME"]
    stdout = shell.exec(f"sqlite3 {database} .dump")
    with open(output_file, "wb") as f:
        f.write(stdout)


def database_exists(connection, database_name):
    return os.access(database_name, os.F_OK)



from django.conf import settings as django_settings
from django.utils.functional import SimpleLazyObject


def _setting(setting, default=None):
    return getattr(django_settings, setting, default)


class Settings:
    def __init__(self):
        self.location = _setting("MIGRATECI_LOCATION", "")
        self.depth = _setting("MIGRATECI_DEPTH", 1)
        self.is_pytest = _setting("MIGRATECI_PYTEST", False)
        self.parallel = _setting("MIGRATECI_PARALLEL", None)

        self.storage_class = _setting(
            "MIGRATECI_STORAGE",
            "django.core.files.storage.FileSystemStorage",
        )


settings = SimpleLazyObject(Settings)



import logging
import os
from subprocess import PIPE, Popen

logger = logging.getLogger(__name__)


class MigrateCIShellException(Exception):
    pass


def exec(command, env=None):
    logger.info(f"Running shell command {command}")

    subprocess_env = os.environ.copy()
    if env:
        subprocess_env.update(env)

    p = Popen(
        command,
        shell=True,
        stdin=PIPE,
        stdout=PIPE,
        stderr=PIPE,
        env=subprocess_env,
    )
    stdout, stderr = p.communicate()
    if stderr:
        raise MigrateCIShellException(stderr.decode())
    return stdout



import logging
from pathlib import Path

from django.core.files.storage import get_storage_class
from django.core.management.base import BaseCommand, CommandError

try:
    from django.test.runner import get_max_test_processes  # type: ignore[attr-defined]
except ImportError:
    # Django<4
    def get_max_test_processes():
        raise CommandError(
            "Django<4 do not implement get_max_test_processes."
            " Use --parallel $(nproc) to not depend on this.",
        )


from django_migrations_ci import django
from django_migrations_ci.conf import settings

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("-n", "--parallel", default=settings.parallel)
        parser.add_argument(
            "--pytest",
            dest="is_pytest",
            action="store_true",
            default=settings.is_pytest,
        )
        parser.add_argument("--location", default=settings.location)
        parser.add_argument(
            "--storage",
            dest="storage_class",
            default=settings.storage_class,
        )
        parser.add_argument("--depth", type=int, default=settings.depth)
        parser.add_argument("--reuse-db", action="store_true", default=False)
        parser.add_argument(
            "--checksum",
            dest="only_print_checksum",
            action="store_true",
            default=False,
            help="Prints the current checksum and exits."
        )

    def _setup_logging(self):
        lib_logger = logging.getLogger("django_migrations_ci")
        lib_logger.setLevel(logging.INFO)
        lib_logger.addHandler(logging.StreamHandler())

    def handle(
        self,
        *args,
        parallel,
        is_pytest,
        location,
        storage_class,
        depth,
        verbosity,
        reuse_db,
        only_print_checksum,
        **options,
    ):
        self._setup_logging()

        if parallel == "auto":
            parallel = get_max_test_processes()
        elif parallel is not None:
            parallel = int(parallel)

        storage_class = get_storage_class(storage_class)
        default_storage_class = get_storage_class(
            "django.core.files.storage.FileSystemStorage",
        )
        if issubclass(storage_class, default_storage_class):
            if not location:
                location = "~/.migrateci"
            location_path = Path(location).expanduser()
            location_path.mkdir(parents=True, exist_ok=True)
            location = str(location_path)

        if verbosity >= 2:
            logger.info(f"Using storage {storage_class=} on {location=}.")
        storage = storage_class(location=location)

        _, files = storage.listdir("")
        files = set(files)
        if verbosity >= 3:
            logger.info(f"Files in storage: {files}")

        unique_connections = django.get_unique_connections()

        current_checksum = None
        checksums = django.hash_files(depth)

        for cached_checksum in checksums:
            # If run with --checksum, print the current checksum and exit.
            if only_print_checksum:
                self.stdout.write(cached_checksum)
                return 0

            # Current checksum is the first result returned from hash_files.
            if current_checksum is None:
                current_checksum = cached_checksum
                if verbosity:
                    logger.info(f"Migrations current checksum is {current_checksum}.")

            cached_files = {
                connection.alias: _migration_filename(connection, cached_checksum)
                for connection in unique_connections
            }
            if all(f in files for f in cached_files.values()):
                if verbosity:
                    logger.info(
                        f"Migrations cache found with checksum {cached_checksum}.",
                    )
                if verbosity >= 3:
                    for other_checksum in checksums:
                        logger.info(
                            f"Calculated checksum {other_checksum} not evaluated.",
                        )
                break

            if verbosity >= 2:
                logger.info(
                    f"Migrations cache NOT found for checksum {cached_checksum}.",
                )

        else:
            cached_checksum = None
            cached_files = None
            if verbosity:
                logger.info("Database cache does not exist.")

        for connection in unique_connections:
            database_name, db_created = django.create_test_db(
                connection,
                verbosity=verbosity,
                keepdb=reuse_db,
            )
            if cached_files and db_created:
                cached_file = cached_files[connection.alias]
                with django.test_db(connection):
                    django.load(
                        connection,
                        cached_file,
                        storage,
                        verbosity=verbosity,
                    )
            elif verbosity >= 2:
                logger.info(f"Reusing database {database_name}.")

        if current_checksum != cached_checksum:
            if verbosity >= 2:
                logger.info(
                    f"Setup test db from {cached_checksum=} to {current_checksum=}.",
                )

            django.setup_test_db(verbosity=verbosity)
            for connection in unique_connections:
                current_file = _migration_filename(connection, current_checksum)
                with django.test_db(connection):
                    django.dump(connection, current_file, storage, verbosity=verbosity)

        if parallel:
            if verbosity >= 2:
                logger.info(f"Clone test db for {parallel=}.")
            for connection in unique_connections:
                with django.test_db(connection):
                    django.clone_test_db(
                        connection=connection,
                        parallel=parallel,
                        is_pytest=is_pytest,
                        verbosity=verbosity,
                    )


def _migration_filename(connection, checksum):
    return f"migrateci-{connection.vendor}-{connection.alias}-{checksum}"
